/*
 * Copyright (c) 2016-2017, Parallels International GmbH
 *
 * This file is part of disp-helper. It's free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 * Our contact details: Parallels International GmbH, Vordergasse 59, 8200
 * Schaffhausen, Switzerland.
 */

#include "Client.hpp"
#include <boost/function.hpp>
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/discrete_distribution.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <sys/stat.h>
#include <dlfcn.h>

#include <cstdlib>

namespace Client
{

namespace
{

std::vector<std::string> GetLogsToCollect()
{
	std::string l(Utils::GlobalSettings().get<std::string>("logs_to_collect", ""));
	if (l.empty())
		return std::vector<std::string>();
	std::vector<std::string> vct;
	boost::split(vct, l, boost::is_any_of(";"), boost::token_compress_on);
	return vct;
}

std::string BuilderFilename()
{
	return "infra_builder.py";
}

bool IsScript(fs::directory_entry const& entry_)
{
	if (!fs::is_regular_file(entry_.status()))
		return false;
	if (!(entry_.status().permissions() & S_IXUSR))
		return false;
	return entry_.path().extension() == ".py";
}

bool IsInfrastructure(fs::directory_entry const& entry_)
{
	return boost::starts_with(entry_.path().filename().string(), "infra_");
}

bool IsPerfomanceScriptPrefix(fs::directory_entry const& entry_)
{
	return boost::starts_with(entry_.path().filename().string(), "perf_");
}

bool IsOrdinaryScript(fs::directory_entry const& entry_)
{
	return IsScript(entry_) && !IsInfrastructure(entry_) && !IsPerfomanceScriptPrefix(entry_);
}

bool IsPerfomanceScript(fs::directory_entry const& entry_)
{
	return IsScript(entry_) && !IsInfrastructure(entry_) && IsPerfomanceScriptPrefix(entry_);
}

void EmptyDummy()
{
}

} // namespace

struct Balancer
{
	Balancer(ba::io_service& service_)
		: m_service(service_),
			m_gen(),
			m_dist(0, 0),
			m_mode(Utils::GlobalSettings().get<std::string>("balancer_mode", "default")),
			m_startDelay(pt::seconds(0)),
			m_actionDelay(pt::seconds(0))
	{
	}

	void Setup(std::time_t lastReport_, fs::path const& scriptsDir_)
	{
		pt::ptime now(pt::second_clock::universal_time());
		pt::time_duration reportPeriod(Utils::Datetime::FromString(
					Utils::GlobalSettings().get<std::string>("report_period", "7d")));
		pt::ptime lastReport(pt::from_time_t(lastReport_));

		pt::ptime toSendReport(now + reportPeriod);
		if (now - lastReport < reportPeriod)
			toSendReport = lastReport + reportPeriod;

		long scriptsNumber(std::count_if(fs::directory_iterator(scriptsDir_),
				fs::directory_iterator(), IsOrdinaryScript));

		// For sufficient delay smoothering before and after sending
		long actionsCount = scriptsNumber + 2;

		if ("off" == m_mode) {
			m_startDelay = toSendReport - now;
			m_actionDelay = pt::seconds(0);
		}

		if ("last_day" == m_mode) {
			m_startDelay = pt::seconds(0);
			if (reportPeriod > pt::hours(24)) {
				m_startDelay = (reportPeriod - pt::hours(24)) - (now - lastReport);
			}
			m_actionDelay = (toSendReport - (now + m_startDelay)) / actionsCount;
		}
		if ("default" == m_mode) {
			m_startDelay = pt::seconds(0);
			m_actionDelay = (toSendReport - now) / actionsCount;
		}
		m_dist = boost::random::uniform_int_distribution<>(
				-m_actionDelay.total_seconds() / 10, m_actionDelay.total_seconds() / 10);
	}

	void StartDelay()
	{
		if ("default" == m_mode)
			return;
		ba::deadline_timer(m_service, m_startDelay).wait();
	}

	void ActionDelay()
	{
		if ("off" == m_mode)
			return;
		pt::time_duration delay = pt::seconds(m_dist(m_gen)) + m_actionDelay;
		ba::deadline_timer(m_service, delay).wait();
	}

private:
	ba::io_service& m_service;
	boost::mt19937 m_gen;
	boost::random::uniform_int_distribution<> m_dist;

	std::string const m_mode;
	pt::time_duration m_startDelay;
	pt::time_duration m_actionDelay;
};

typedef int (*sender_t)(const char*, const char*, char**);
typedef const char* (*protocol_t)(void);
Collector::Collector(ba::io_service& service_)
	: m_service(service_),
		m_timer(m_service, pt::seconds(1)),
		m_perfomance_timer(m_service, pt::seconds(1)),
		m_scriptsDir("/usr/share/virtuozzo/cep-scripts"),
		m_preliminaryDir("/vz/tmp/cep/output/preliminary"),
		m_finalResultsDir("/vz/tmp/cep/output/final"),
		m_archive("/vz/tmp/report.tar.gz"),
		m_flagFile("/vz/tmp/cep/.sent"),
		m_balancer(new Balancer(m_service))
{
	fs::directory_iterator last;
	for (fs::directory_iterator it("/usr/lib64/disp-helper-plugins"); it != last; ++it) {
		try {
			if (it->path().extension() != ".so")
				continue;
			::dlerror();
			void* handle = ::dlopen(it->path().string().c_str(), RTLD_LAZY);
			if (NULL == handle) {
				LOG_ERROR_FMT(boost::format("Unable to open shared library %s %s") % it->path().string() % ::dlerror());
				continue;
			}

			protocol_t protocol((protocol_t)::dlsym(handle, "protocol"));
			const char* error(::dlerror());
			if (NULL != error) {
				LOG_ERROR_FMT(boost::format("Unable to load symbols from %s (%s)")
						% it->path().string() % error);
				continue;
			}

			sender_t sender((sender_t)::dlsym(handle, "sender"));
			error = ::dlerror();
			if (NULL != error) {
				LOG_ERROR_FMT(boost::format("Unable to load symbols from %s (%s)")
						% it->path().string() % error);
				continue;
			}
			m_senders[protocol()] = sender;
		} catch (std::exception const& exc) {
			LOG_ERROR_FMT(boost::format("Exception %s") % exc.what());
			exit(EXIT_FAILURE);
		}
	}
}

bool Collector::ExecuteScripts(boost::function<bool (fs::directory_entry const&)> predicate_,
		boost::function<void ()> delay_) const
{
	fs::directory_iterator last;
	for (fs::directory_iterator it(m_scriptsDir); it != last; ++it) {
		try {
			if (!predicate_(*it))
				continue;

			char **args(NULL);
			try {
				args = static_cast<char**>(::calloc(5, sizeof(char*)));
				args[0] = ::strdup("/usr/bin/python");
				args[1] = ::strdup(fs::absolute(it->path()).string().c_str());
				args[2] = static_cast<char*>(::malloc(128));
				::snprintf(args[2], 128, "%s", m_preliminaryDir.string().c_str());
				args[3] = static_cast<char*>(::malloc(128));
				::snprintf(args[3], 128, "--time=%lu", Utils::Datetime::CurrentTimestamp());
				args[4] = NULL;
				Utils::Base::Instance().IoService().dispatch(boost::bind(Utils::SafeExecute, args, delay_));
				for (size_t i = 0; i < 5; ++i)
					::free(args[i]);
				free(args);
			} catch (std::exception const& exc) {
				if (args) {
					for (size_t i = 0; i < 5; ++i)
						::free(args[i]);
					free(args);
				}
			}
		} catch (std::exception const& e) {
			LOG_ERROR_FMT(boost::format("%s %s") % it->path().filename() % e.what());
			return false;
		}
	}
	return true;
}

void Collector::PrepareFilesystem() const
{
	try {
		boost::system::error_code ec;
		fs::remove_all(m_preliminaryDir, ec);
		fs::remove_all(m_finalResultsDir, ec);
		if (Utils::GlobalSettings().get<std::string>("debug_results", "off") != "on")
			fs::remove(m_archive, ec);
		fs::create_directories(m_preliminaryDir, ec);
		fs::create_directories(m_finalResultsDir, ec);
	} catch (fs::filesystem_error& e) {
		LOG_ERROR_FMT(boost::format("Filesystem error: %s") % e.what());
		exit(EXIT_FAILURE);
	}

	pt::ptime now(pt::second_clock::universal_time());
	if (!fs::exists(m_flagFile) ||
		pt::from_time_t(fs::last_write_time(m_flagFile)) > now)
		Utils::TouchFile(m_flagFile);
}

bool Collector::PrepareResults()
{
	fs::path builder(m_scriptsDir / BuilderFilename());

	char **args(NULL);
	try {
		args = static_cast<char**>(::calloc(6, sizeof(char*)));
		args[0] = ::strdup("/usr/bin/python");
		args[1] = ::strdup(builder.string().c_str());
		args[2] = ::strdup(m_preliminaryDir.string().c_str());
		args[3] = static_cast<char*>(::malloc(128));
		::snprintf(args[3], 128, "--time=%lu", Utils::Datetime::CurrentTimestamp());
		args[4] = ::strdup("/vz/tmp/cep/output/final/result.json");
		args[5] = NULL;
		Utils::SafeExecute(args, EmptyDummy);
	} catch (std::exception const& exc) {
		if (args) {
			for (size_t i = 0; i < 5; ++i)
				::free(args[i]);
			free(args);
		}
	}

	std::vector<std::string> vct(GetLogsToCollect());
	for (size_t i = 0; i < vct.size(); ++i) {
		if (!fs::exists(vct[i])) {
			LOG_WARN_FMT(boost::format("Unable to find file %s, please check settings") % vct[i]);
			continue;
		}
		fs::path p(vct[i]);
		fs::copy_file(p, m_finalResultsDir / p.filename(),
			  fs::copy_option::overwrite_if_exists);
	}

	// TODO use libtar
	std::ostringstream ostr;
	ostr << "/usr/bin/tar -zcf " << m_archive << " --directory=\""
		<< fs::absolute(m_finalResultsDir) << "\" .";
	return !::system(ostr.str().c_str());
}

void Collector::Iteration()
{
	PrepareFilesystem();
	m_balancer->Setup(fs::last_write_time(m_flagFile), m_scriptsDir);
	m_balancer->StartDelay();
	ExecuteScripts(boost::bind(IsOrdinaryScript, _1), boost::bind(&Balancer::ActionDelay, m_balancer));
	PrepareResults();

	m_balancer->ActionDelay();
	m_service.post(boost::bind(&Collector::SendResults, this));
}

void Collector::PerfomanceIteration()
{
	ExecuteScripts(boost::bind(IsPerfomanceScript, _1), EmptyDummy);
	pt::time_duration frequency(Utils::Datetime::FromString(Utils::GlobalSettings().get<std::string>("frequency_period", "5m")));
	m_perfomance_timer.expires_from_now(frequency);
	m_perfomance_timer.async_wait(boost::bind(&Collector::PerfomanceIteration, this));
}

void Collector::Run()
{
	Utils::Base::Instance().IoService().dispatch(boost::bind(&Collector::PerfomanceIteration, this));
	Utils::Base::Instance().IoService().dispatch(boost::bind(&Collector::Iteration, this));
}

void Collector::SendResults()
{
	std::string protocol(Utils::GlobalSettings().get<std::string>("protocol", "CURL"));
	typename senders_t::const_iterator it(m_senders.find(protocol));
	if (m_senders.end() == it) {
		LOG_ERROR_FMT(boost::format("Unable to find sender for '%s' protocol") % protocol);
		::exit(EXIT_FAILURE);
	}
	char *answer(NULL);
	std::string server(Utils::GlobalSettings().get<std::string>("master_server",
		"https://report.virtuozzo.com/post"));
	int error(it->second(m_archive.string().c_str(), server.c_str(), &answer));
	if (0 == error) {
		Utils::TouchFile(m_flagFile);

		std::vector<std::string> vct(GetLogsToCollect());
		for (size_t i = 0; i < vct.size(); ++i) {
			if (!fs::exists(vct[i]))
				continue;
			std::ofstream ofs;
			ofs.open(vct[i].c_str(), std::ofstream::out | std::ofstream::trunc);
			ofs.close();
		}

		LOG_INFO_FMT(boost::format("Successfully sent report; Server response: %s") % answer);
		m_balancer->ActionDelay();
		m_service.post(boost::bind(&Collector::Iteration, this));
	} else {
		std::string s(Utils::GlobalSettings().get<std::string>("retry_period", "2h"));
		LOG_ERROR_FMT(boost::format("Unable to send. Sender return %d, "
					"retrying send attempt in %s") % error % s);
		pt::time_duration retryPeriod(Utils::Datetime::FromString(s));
		m_timer.expires_from_now(retryPeriod);
		m_timer.async_wait(boost::bind(&Collector::SendResults, this));
	}
	::free(answer);
}

} // namespace Client
