/*
 * Copyright (c) 2016-2017, Parallels International GmbH
 *
 * This file is part of disp-helper. It's free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 * Our contact details: Parallels International GmbH, Vordergasse 59, 8200
 * Schaffhausen, Switzerland.
 */

#include "Utils.hpp"
#include <iostream>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/file.h>

#include <boost/program_options.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <boost/property_tree/xml_parser.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/algorithm/string/classification.hpp>


namespace Utils
{

namespace Details
{

bool Daemonize()
{
	if (Utils::Base::Instance().Settings().get<bool>("interactive", false)) {
		std::cout << "Collector is working in interactive mode" << std::endl;
		return true;
	}
	Base::Instance().IoService().notify_fork(ba::io_service::fork_prepare);
	pid_t p(::fork());
	if (p > 0) {
		Base::Instance().IoService().notify_fork(ba::io_service::fork_parent);
		exit(EXIT_SUCCESS);
	}
	if (-1 == p) {
		LOG_ERROR("Unable to fork");
		exit(EXIT_FAILURE);
	}
	Base::Instance().IoService().notify_fork(ba::io_service::fork_child);
	if (-1 == setsid()) {
		LOG_ERROR("Unable to setsid");
		return false;
	}
	::signal(SIGCHLD, SIG_IGN);
	::signal(SIGHUP, SIG_IGN);
	if (-1 == ::chdir("/")) {
		LOG_ERROR("Unable to chdir");
		return false;
	}
	Base::Instance().IoService().notify_fork(ba::io_service::fork_prepare);
	p = ::fork();
	if (p > 0) {
		Base::Instance().IoService().notify_fork(ba::io_service::fork_parent);
		exit(EXIT_SUCCESS);
	}
	Base::Instance().IoService().notify_fork(ba::io_service::fork_child);
	if (-1 == p) {
		LOG_ERROR("Unable to fork");
		exit(EXIT_FAILURE);
	}
	::umask(027);
	int fd(::open("/dev/null", O_RDWR));
	if (-1 == fd) {
		LOG_ERROR("Unable to open /dev/null");
		return false;
	}
	::dup2(fd, STDIN_FILENO);
	::dup2(fd, STDOUT_FILENO);
	::dup2(fd, STDERR_FILENO);
	::close(fd);
	signal(SIGCHLD, SIG_DFL);
	return true;
}

bool IsVzLicDaemonRunning()
{
	namespace fs = boost::filesystem;
	fs::path vzlicmon("/usr/sbin/vzlicmonitor");
	fs::directory_iterator last;
	for (fs::directory_iterator it("/proc"); it != last; ++it) {
		fs::path p(it->path() / "exe");
		if (fs::is_regular_file(p) && fs::read_symlink(p) == vzlicmon)
				return true;
	}
	return false;
}

bool ReadDispUuid()
{
	boost::property_tree::ptree p;
	try {
		boost::property_tree::read_xml("/etc/vz/dispatcher.xml", p);
		std::string uuid(p.get<std::string>("ParallelsPreferences.ServerIdentification.ServerUuid"));
		boost::trim_if(uuid, boost::is_any_of("{}"));
		LOG_INFO("Dispatcher uuid: " + uuid);
	} catch (std::exception const& exc) {
		LOG_ERROR(exc.what());
		return false;
	}
	return true;
}

bool PreparePidFile()
{
	int fd(-1);
	fd = ::open("/var/run/disp_helper.pid", O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR);
	if (-1 == fd)
		return false;
    while (0 != ::flock(fd, LOCK_EX | LOCK_NB))
    {
        if (EWOULDBLOCK == errno)
			return false;
        if (EINTR != errno)
			return false;
    }
	char buf[1024];
	::snprintf(buf, sizeof(buf), "%ld\n", ::getpid());
	if (-1 == ::write(fd, buf, ::strlen(buf) + 1))
		return false;
	return true;
}

} // namespace Details

namespace Datetime
{

uint64_t CurrentTimestamp()
{
	return to_time_t(pt::second_clock::universal_time());
}

std::string CurrentTime()
{
	return to_simple_string(pt::second_clock::universal_time());
}

pt::time_duration FromString(std::string const& str)
{
	std::istringstream is(str);
	unsigned long quantity;
	char units;
	is >> quantity >> units;
	switch (units) {
		case 's':
			return pt::seconds(quantity);
			break;
		case 'm':
			return pt::minutes(quantity);
			break;
		case 'h':
		case 'H':
			return pt::hours(quantity);
			break;
		case 'd':
		case 'D':
			return pt::hours(quantity * 24);
			break;
		case 'w':
		case 'W':
			return pt::hours(quantity * 24 * 7);
			break;
		case 'M':
			return pt::hours(quantity * 24 * 7 * 30);
			break;
		default:
			throw std::invalid_argument("Bad time duration unit");
			break;
	}
}

} // namespace Datetime

bool Base::LoadSettings(std::string const& path)
{
	try {
		boost::property_tree::read_json(path, m_settings);
		return true;
	} catch (std::exception const& exc) {
		LOG_ERROR(exc.what());
		return false;
	}
}

bool Init(int argc, char *argv[])
{
	using namespace boost::program_options;
	std::string mainserver;
	std::string slaveserver;
	std::string settingsfile;

	options_description desc("Allowed options");
	desc.add_options()
	("help,h", "Shows this help")
	("main-server", value<std::string>(&mainserver)->default_value("stats.openvz.org:9898"), "")
	("slave-server", value<>(&slaveserver)->default_value("key-admin.address.com:9898"),
		 "Secondary server to collect reports, which is used in case of master unavailability")
	("one-file-mode,o", "Send whole report as one file")
	("settings,s", value<std::string>(&settingsfile)->default_value("/etc/vz/disp_helper.json"), "")
	("interactive,i", "run in interactive mode, not as daemon")
	;

	variables_map vm;
	store(parse_command_line(argc, argv, desc), vm);

	if (vm.count("help")) {
		std::cout << desc << std::endl;
		::exit(EXIT_SUCCESS);
	}

	if (!Base::Instance().LoadSettings(vm["settings"].as<std::string>())) {
		std::cerr << "Unable to read settings" << std::endl;
		return false;
	}
	if (vm.count("interactive"))
		Base::Instance().Settings().put("interactive", true);
	if (vm.count("one-file-mode"))
		Base::Instance().Settings().put("one-file-mode", true);

	return Base::Instance().Prepare();
}

bool TouchFile(fs::path const& file_)
{
    int fd(::open(file_.string().c_str(), O_WRONLY|O_CREAT|O_NOCTTY|O_NONBLOCK, 0666));
    if (-1 == fd)
		return false;
    bool r(-1 != ::futimens(fd, NULL));
	::close(fd);
	return r;
}

bool SafeExecute(char *args[], boost::function<void ()> delay_)
{
	if (args == NULL) {
		LOG_ERROR("Empty arguments");
		return false;
	}

	std::stringstream ss;
	char **t(&args[0]);
	for (; *t; ++t)
		ss << " " << *t;

	LOG_INFO_FMT(boost::format("Execute %s") % ss.str());

	Base::Instance().IoService().notify_fork(ba::io_service::fork_prepare);
	pid_t pid(::fork());
	if (pid < 0) {
		LOG_ERROR("Unable to fork");
		return false;
	}

	if (pid == 0) {
		Base::Instance().IoService().notify_fork(ba::io_service::fork_child);
		if (::execv(args[0], args) == -1) {
			LOG_ERROR("Unable to execve");
			::exit(1);
		}
	} else {
		Base::Instance().IoService().notify_fork(ba::io_service::fork_parent);
		uint64_t t(Utils::Datetime::CurrentTimestamp());
		int stat, ret;
		while ((ret = ::waitpid(pid, &stat, WNOHANG)) == 0) {
			if (Utils::Datetime::CurrentTimestamp() - t > 10 * 60) {
				LOG_INFO("Timeout");
				::kill(pid, SIGKILL);
				ret = -1;
				break;
			}
			ba::deadline_timer(Base::Instance().IoService(), pt::milliseconds(10)).wait();
		}
		if (ret == -1) {
			std::string res(args[0]);
			for (char** a = &args[1]; *a; ++a)
				res = res + " " + *a;
			LOG_WARN_FMT(boost::format("Execution error (%s): %s") % strerror(errno) % res);
			return false;
		}
	}
	delay_();
	return true;
}

} // namespace Utils
