/*
 * Copyright (c) 2016-2017, Parallels International GmbH
 *
 * This file is part of disp-helper. It's free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 * Our contact details: Parallels International GmbH, Vordergasse 59, 8200
 * Schaffhausen, Switzerland.
 */

#ifndef UTILS_HPP_
#define UTILS_HPP_

#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <map>

#include <boost/asio.hpp>
#include <boost/thread/once.hpp>
#include <boost/thread/thread.hpp>
#include <boost/noncopyable.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#include <boost/cstdint.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/function.hpp>

#include <unistd.h>
#include <systemd/sd-daemon.h>

namespace fs = boost::filesystem;
namespace pt = boost::posix_time;
namespace ba = boost::asio;

namespace Utils
{

namespace Datetime
{

inline time_t to_time_t(pt::ptime const& a_posix_time)
{
	pt::ptime const time_since_epoch(boost::gregorian::date(1970, 1, 1));
	return pt::time_duration(a_posix_time - time_since_epoch).total_seconds();
}

uint64_t CurrentTimestamp();
std::string CurrentTime();
pt::time_duration FromString(std::string const& str);

} // namespace Datetime


template <typename T>
class Singleton: boost::noncopyable
{
public:
    static T& Instance()
	{
        boost::call_once(Singleton<T>::m_init_flag, Init);
        return *m_obj;
    }

private:
    static void Init()
	{
        m_obj = new T;
        std::atexit(Destroy);
    }

    static void Destroy()
	{
        boost::call_once(Singleton<T>::m_delete_flag, DeleteObj);
    }

    static void DeleteObj()
	{
        delete m_obj;
    }

private:
    static T* m_obj;
    static boost::once_flag m_init_flag;
    static boost::once_flag m_delete_flag;
};

template <typename T>
T* Singleton<T>::m_obj = 0;

#ifdef BOOST_THREAD_PROVIDES_ONCE_CXX11
template <typename T>
boost::once_flag Singleton<T>::m_init_flag;

template <typename T>
boost::once_flag Singleton<T>::m_delete_flag;
#else
template <typename T>
boost::once_flag Singleton<T>::m_init_flag = BOOST_ONCE_INIT;

template <typename T>
boost::once_flag Singleton<T>::m_delete_flag = BOOST_ONCE_INIT;
#endif

namespace Details
{

class BasicLogger: private boost::noncopyable
{
public:
	explicit BasicLogger(ba::io_service& service_, std::string const& id_, std::ofstream& file_)
		: m_service(service_), m_id(id_), m_file(file_)
	{
	}

	~BasicLogger()
	{
	}

	void Log(std::string const& message_)
	{
		std::ostringstream os;
		os << Datetime::CurrentTime() << " (" << boost::this_thread::get_id()
			<< ") " << m_id << ": " << message_;
		m_service.dispatch(boost::bind(&BasicLogger::LogImpl, this, os.str()));
	}

private:
	void LogImpl(std::string const& text_)
	{
		m_file << text_ << std::endl;
	}

private:
	ba::io_service& m_service;
	std::string m_id;
	std::ofstream& m_file;
};

struct Log
{
	Log(ba::io_service& service_)
		: m_info(service_, "I", m_file),
		m_error(service_, "E", m_file),
		m_warning(service_, "W", m_file)
	{
	}

	BasicLogger& Info()
	{
		return m_info;
	}

	BasicLogger& Error()
	{
		return m_error;
	}

	BasicLogger& Warning()
	{
		return m_warning;
	}

	void Prepare(bool interactive_, std::string const& file_)
	{
		if (interactive_) {
			m_file.basic_ios<char>::rdbuf(std::cout.rdbuf());
			return;
		}

		m_file.close();
		m_file.clear();
		m_file.open(file_.c_str(), std::ios_base::out | std::ios_base::trunc);
	}

private:
	std::ofstream m_file;
	BasicLogger m_info;
	BasicLogger m_error;
	BasicLogger m_warning;
};

bool Daemonize();

bool IsVzLicDaemonRunning();

bool ReadDispUuid();

bool PreparePidFile();

} // namespace Details

struct Base: Singleton<Base>
{
	Base()
		: m_io_service(),
			m_log(m_io_service),
			m_settings()
	{
	}

	bool LoadSettings(std::string const& path);

	boost::property_tree::ptree& Settings()
	{
		return m_settings;
	}

	ba::io_service& IoService()
	{
		return m_io_service;
	}

	bool Prepare()
	{
		m_log.Prepare(m_settings.get<bool>("interactive", false),
			m_settings.get<std::string>("logfile", "/var/log/collector.log"));
		if (!Utils::Details::Daemonize())
			return false;

		if (!Utils::Details::PreparePidFile())
			return false;
		::sd_notify(0, "READY=1");
		return true;
	}

	void Run()
	{
		boost::thread_group tg;
		for (size_t i = 0; i < 10; ++i) // TODO think about thread count
			tg.create_thread(boost::bind(&ba::io_service::run, &m_io_service));
		tg.join_all();
	}

	Details::Log& Log()
	{
		return m_log;
	}

private:
	ba::io_service m_io_service;
	Details::Log m_log;
	boost::property_tree::ptree m_settings;
};


bool Init(int argc, char *argv[]);

inline boost::property_tree::ptree& GlobalSettings()
{
	return Utils::Base::Instance().Settings();
}

#define LOG_WARN(msg) do { Utils::Base::Instance().Log().Warning().Log(msg); } while(0)
#define LOG_INFO(msg) do { Utils::Base::Instance().Log().Info().Log(msg); } while(0)
#define LOG_ERROR(msg) do { Utils::Base::Instance().Log().Error().Log(msg); } while(0)
#define LOG_WARN_FMT(msg) do { Utils::Base::Instance().Log().Warning().Log((msg).str()); } while(0)
#define LOG_INFO_FMT(msg) do { Utils::Base::Instance().Log().Info().Log((msg).str()); } while(0)
#define LOG_ERROR_FMT(msg) do { Utils::Base::Instance().Log().Error().Log((msg).str()); } while(0)

bool TouchFile(fs::path const& file_);
bool SafeExecute(char *args[], boost::function<void()> delay_);

} // namespace Utils

#endif
